@@ -2,43 +2,17 @@
 //! them to cover the full gamut of POSIX file types, and wrapping them in order to improve the
 //! usefulness of error messages by providing additional context.
 
+use crate::error::{Error, Result};
 use nix::sys::stat::Mode;
 use nix::unistd;
 use std::convert::TryInto;
-use std::error::Error as BaseError;
-use std::fmt;
-use std::fs::{self, DirBuilder, File, Metadata, OpenOptions, Permissions, ReadDir};
+use std::fs::{self, DirBuilder, DirEntry, File, Metadata, OpenOptions, Permissions, ReadDir};
 use std::os::unix::fs::{self as unix, DirBuilderExt, FileTypeExt, OpenOptionsExt, PermissionsExt};
 use std::path::{Path, PathBuf};
 
-#[derive(Debug)]
-pub struct Error {
-    message: String,
-}
-
-impl fmt::Display for Error {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}", self.message)
-    }
-}
-
-impl<T: BaseError> From<T> for Error {
-    fn from(other: T) -> Self {
-        Error {
-            message: other.to_string(),
-        }
-    }
-}
-
-impl Error {
-    pub fn new(message: String) -> Self {
-        Error { message }
-    }
-}
-
 macro_rules! wrap {
     ($namespace:ident, $function:ident, $payload:ty) => {
-        pub fn $function<P: AsRef<Path>>(path: P) -> Result<$payload, Error> {
+        pub fn $function<P: AsRef<Path>>(path: P) -> Result<$payload> {
             $namespace::$function(path.as_ref())
                 .map_err(|err| Error::new(format!("{}: {}", path.as_ref().display(), err)))
         }
@@ -47,10 +21,7 @@ macro_rules! wrap {
 
 macro_rules! wrap2 {
     ($function:ident, $namespace:ident, $payload:ty) => {
-        pub fn $function<P: AsRef<Path>, Q: AsRef<Path>>(
-            source: P,
-            dest: Q,
-        ) -> Result<$payload, Error> {
+        pub fn $function<P: AsRef<Path>, Q: AsRef<Path>>(source: P, dest: Q) -> Result<$payload> {
             let (source, dest) = (source.as_ref(), dest.as_ref());
             $namespace::$function(source, dest).map_err(|err| {
                 Error::new(format!("{}, {}: {}", source.display(), dest.display(), err))
@@ -77,7 +48,14 @@ macro_rules! make_error_message {
     };
 }
 
-pub fn create_dir<P: AsRef<Path>>(path: P, mode: u32) -> Result<(), Error> {
+pub fn entry_file_type(entry: &DirEntry) -> Result<FileType> {
+    match entry.file_type() {
+        Err(err) => Err(Error::new(format!("{}: {}", entry.path().display(), err))),
+        Ok(file_type) => Ok(FileType::from(file_type)),
+    }
+}
+
+pub fn create_dir<P: AsRef<Path>>(path: P, mode: u32) -> Result<()> {
     let path = path.as_ref();
     DirBuilder::new()
         .mode(mode)
@@ -85,7 +63,7 @@ pub fn create_dir<P: AsRef<Path>>(path: P, mode: u32) -> Result<(), Error> {
         .map_err(make_error_message!(path))
 }
 
-pub fn create<P: AsRef<Path>>(path: P, mode: u32) -> Result<File, Error> {
+pub fn create<P: AsRef<Path>>(path: P, mode: u32) -> Result<File> {
     let path = path.as_ref();
     OpenOptions::new()
         .mode(mode)
@@ -96,7 +74,7 @@ pub fn create<P: AsRef<Path>>(path: P, mode: u32) -> Result<File, Error> {
         .map_err(make_error_message!(path))
 }
 
-pub fn mkfifo<P: AsRef<Path>>(path: P, permissions: Permissions) -> Result<(), Error> {
+pub fn mkfifo<P: AsRef<Path>>(path: P, permissions: Permissions) -> Result<()> {
     let path = path.as_ref();
     let mode = Mode::from_bits_truncate(permissions.mode().try_into()?);
     unistd::mkfifo(path, mode).map_err(make_error_message!(path))
@@ -105,35 +83,36 @@ pub fn mkfifo<P: AsRef<Path>>(path: P, permissions: Permissions) -> Result<(), E
 #[derive(Debug)]
 pub enum FileType {
     Regular,
-    Directory(Metadata),
+    Directory,
     Symlink,
-    Fifo(Metadata),
+    Fifo,
     Socket,
-    CharacterDevice(Metadata),
-    BlockDevice(Metadata),
+    CharacterDevice,
+    BlockDevice,
 }
 
-pub fn file_type(path: &Path) -> Result<FileType, Error> {
-    let metadata = symlink_metadata(path)?;
-    let file_type = metadata.file_type();
-    Ok(if file_type.is_file() {
-        FileType::Regular
-    } else if file_type.is_dir() {
-        FileType::Directory(metadata)
-    } else if file_type.is_symlink() {
-        FileType::Symlink
-    } else if file_type.is_fifo() {
-        FileType::Fifo(metadata)
-    } else if file_type.is_socket() {
-        FileType::Socket
-    } else if file_type.is_char_device() {
-        FileType::CharacterDevice(metadata)
-    } else if file_type.is_block_device() {
-        FileType::BlockDevice(metadata)
-    } else {
-        unreachable!(
-            "{}: file appears to exist but is an unknown type",
-            path.display()
-        );
-    })
+impl From<std::fs::FileType> for FileType {
+    fn from(file_type: std::fs::FileType) -> Self {
+        if file_type.is_file() {
+            FileType::Regular
+        } else if file_type.is_dir() {
+            FileType::Directory
+        } else if file_type.is_symlink() {
+            FileType::Symlink
+        } else if file_type.is_fifo() {
+            FileType::Fifo
+        } else if file_type.is_socket() {
+            FileType::Socket
+        } else if file_type.is_char_device() {
+            FileType::CharacterDevice
+        } else if file_type.is_block_device() {
+            FileType::BlockDevice
+        } else {
+            unreachable!("file appears to exist but is an unknown type",);
+        }
+    }
+}
+
+pub fn file_type(path: &Path) -> Result<FileType> {
+    Ok(FileType::from(symlink_metadata(path)?.file_type()))
 }
