@@ -4,16 +4,17 @@ use std::collections::HashMap;
 use std::env;
 use std::ffi::OsStr;
 use std::fmt::Display;
-use std::fs::Metadata;
 use std::io;
 use std::ops::BitOr;
 use std::os::unix::fs::{MetadataExt, PermissionsExt};
 use std::path::{Path, PathBuf};
 use std::process;
 
+pub mod error;
 pub mod filesystem;
 
-use crate::filesystem::{self as fs, Error, FileType};
+use crate::error::{Error, Result};
+use crate::filesystem::{self as fs, FileType};
 
 pub fn fatal(message: impl Display) -> ! {
     eprintln!("{}", message);
@@ -26,22 +27,23 @@ pub fn fatal(message: impl Display) -> ! {
 // long-running jobs) as opposed to propagating it upwards and printing all errors at the end.
 // However, at the end of the process we still need to know whether or not an error occurred at any
 // point in order to set the exit code appropriately.
-fn copy_file(source: &Path, dest: &Path) -> bool {
-    fn __copy_file(source: &Path, dest: &Path) -> Result<bool, Error> {
-        match fs::file_type(source)? {
+fn copy_file(source: &Path, source_type: Result<FileType>, dest: &Path) -> bool {
+    fn __copy_file(source: &Path, source_type: Result<FileType>, dest: &Path) -> Result<bool> {
+        match source_type? {
             FileType::Regular => {
                 fs::copy(source, dest)?;
             }
-            FileType::Directory(metadata) => return copy_directory((source, metadata), dest),
+            FileType::Directory => return copy_directory(source, dest),
             FileType::Symlink => fs::symlink(fs::read_link(source)?, dest)?,
-            FileType::Fifo(metadata) => fs::mkfifo(dest, metadata.permissions())?,
+            FileType::Fifo => fs::mkfifo(dest, fs::symlink_metadata(source)?.permissions())?,
             FileType::Socket => {
                 return Err(Error::new(format!(
                     "{}: sockets cannot be copied",
                     source.display(),
                 )));
             }
-            FileType::CharacterDevice(metadata) | FileType::BlockDevice(metadata) => {
+            FileType::CharacterDevice | FileType::BlockDevice => {
+                let metadata = fs::symlink_metadata(source)?;
                 let mut source = fs::open(source)?;
                 let mut dest = fs::create(dest, metadata.permissions().mode())?;
                 io::copy(&mut source, &mut dest)?;
@@ -50,29 +52,34 @@ fn copy_file(source: &Path, dest: &Path) -> bool {
         Ok(false)
     }
 
-    __copy_file(source, dest).unwrap_or_else(|err| {
+    __copy_file(source, source_type, dest).unwrap_or_else(|err| {
         eprintln!("{}", err);
         true
     })
 }
 
-fn copy_directory(source: (&Path, Metadata), dest: &Path) -> Result<bool, Error> {
-    let (source, metadata) = source;
-    fs::create_dir(dest, metadata.permissions().mode())?;
-    Ok(fs::read_dir(source)?
-        .collect::<Box<_>>()
-        .into_par_iter()
-        .map(|entry| match entry {
-            Ok(entry) => copy_file(&entry.path(), &dest.join(entry.file_name())),
+fn copy_directory(source: &Path, dest: &Path) -> Result<bool> {
+    fs::create_dir(dest, fs::symlink_metadata(source)?.permissions().mode())?;
+    let (mut entries, mut has_err) = (Vec::new(), false);
+    for entry in fs::read_dir(source)? {
+        match entry {
+            Ok(entry) => entries.push((entry.file_name(), fs::entry_file_type(&entry))),
             Err(err) => {
                 eprintln!("{}", err);
-                true
+                has_err = true;
             }
+        }
+    }
+    entries.shrink_to_fit();
+    Ok(entries
+        .into_par_iter()
+        .map(|(file_name, file_type)| {
+            copy_file(&source.join(&file_name), file_type, &dest.join(&file_name))
         })
-        .reduce(|| false, BitOr::bitor))
+        .reduce(|| has_err, BitOr::bitor))
 }
 
-fn reject_self_copies(sources: &[PathBuf], dest: &Path) -> Result<(), Error> {
+fn reject_self_copies(sources: &[PathBuf], dest: &Path) -> Result<()> {
     let current_dir = env::current_dir()?;
     let mut prefix = Path::new("");
     // We make `dest` absolute because for relative paths the final non-`None` value returned by
@@ -125,7 +132,7 @@ fn reject_self_copies(sources: &[PathBuf], dest: &Path) -> Result<(), Error> {
     }
 }
 
-fn file_names(sources: &[PathBuf]) -> Result<Vec<&OsStr>, Error> {
+fn file_names(sources: &[PathBuf]) -> Result<Vec<&OsStr>> {
     let source_file_names = sources
         .iter()
         .map(|source| {
@@ -136,7 +143,7 @@ fn file_names(sources: &[PathBuf]) -> Result<Vec<&OsStr>, Error> {
                 ))
             })
         })
-        .collect::<Result<Vec<_>, _>>()?;
+        .collect::<Result<Vec<_>>>()?;
     let mut sources_by_name: HashMap<_, Vec<_>> = HashMap::new();
     for (source, file_name) in sources.iter().zip(&source_file_names) {
         sources_by_name.entry(file_name).or_default().push(source);
@@ -180,10 +187,13 @@ fn copy_into(sources: &[PathBuf], dest: &Path) -> bool {
         .zip(file_names(sources).unwrap_or_else(|err| fatal(err)))
         .collect::<Box<_>>()
         .into_par_iter()
-        .map(|(source, file_name)| copy_file(&source, &dest.join(file_name)))
+        .map(|(source, file_name)| copy_file(&source, fs::file_type(source), &dest.join(file_name)))
         .reduce(|| false, BitOr::bitor)
 }
 
+// The `allow` here is present because clippy doesn't realize that `source` must be of
+// type `&PathBuf` in order for the call to `array::from_ref` to typecheck.
+#[allow(clippy::ptr_arg)]
 fn copy_single(source: &PathBuf, dest: &Path) -> bool {
     let source_metadata = fs::symlink_metadata(source).unwrap_or_else(|err| fatal(err));
     match (fs::metadata(dest), fs::symlink_metadata(dest)) {
@@ -193,7 +203,7 @@ fn copy_single(source: &PathBuf, dest: &Path) -> bool {
             source.display(),
             dest.display()
         )),
-        _ => copy_file(source, dest),
+        _ => copy_file(source, fs::file_type(source), dest),
     }
 }
 
